#include <stdio.h>
#include <string.h>
#include "gpmi.h"

void test_packages(gpmi_module *mod_example);

/* This function is used to print a detailed error message */
void print_gpmi_error(gpmi_err_stack_t *entry, char *string)
{
	printf("-> %s\n", string);
}

int main() {
	gpmi_module *mod_example;
	int n;

	/* Feature: changing default paths
	   Level:   recommended

		 With bad path settings, gpmi won't be able to reach your modules
		 and packages. The library has defaults that may work if you follow
		 its dir structure when creating your directory tree. Now
		 we just set it up so you can run the example from the example/
		 dir from the gpmi source. If you clone this example, you will need
		 to change it or leave it out. When you don't set it, gpmi will use
		 default paths (pointing to the dir where gpmi is installed in)
		 Warning: always assign an allocated string, NULL is not valid.
	*/
	gpmi_path_append(GPMI_PATH_MODULES, GPMI_ROOT "/modules");
#ifdef WINDLL
	gpmi_path_append(GPMI_PATH_PACKAGES, GPMI_ROOT "/gpmi_pkg");
#else
	gpmi_path_append(GPMI_PATH_PACKAGES, GPMI_ROOT "/packages");
#endif

	/* Feature: init the lib
	   Level:   critical
	*/
	gpmi_init();


	/* Feature: enable detailed error tracing
		 Level:   optional
	*/
	gpmi_err_stack_enable();

	/* Feature: loading a library
	   Level:   required
	*/
	mod_example = gpmi_mod_load("example", "Init-params");

	/* checking if suceed
	   Level:   required

		 gpmi will crash for NULL pointers as module handlers!
	*/
	if (mod_example == NULL) {
		printf("Couldn't open example module, error stack:\n");

		/* Feature: printing detailed error trace
		   Level:   optional
		*/
		gpmi_err_stack_process_str(print_gpmi_error);
		return 1;
	}

	/* Feature: safe calling of the debug function of the module
	   Level:   optional

		 Script interpreters *should* pipe the debug command in a safe
		 interpreter to provide a nice way to debug scripts on the fly.
	*/
	gpmi_call_Debug(mod_example, "Debug works.");

	/* Now we just have a main loop for our application that runs 4 times*/
	for(n = 0; n < 4; n++) {
		/* Feature: calling RunTick of the module
		   Level:   optional

			 If your application has a main loop and the module has a RunTick
			 handler and you want the module to run parallell with your app,
			 you should do this. Typical for game AIs, applications where
			 the module needs timers or in general, where modules work async.
			 Unlikely for network clients.
		*/
		gpmi_call_RunTick(mod_example, n);
	}

	/* Feature: calling RunData of the module
	   Level:   optional

		 This is the default way to pass data to a module when you use the
		 module as a filter. Use it if your modules don't need to do async
		 tasks but only process some sort of incoming data. Ideal for
		 servers, data flow processing
	*/
	gpmi_call_RunData(mod_example, (void *)0x1234);


	test_packages(mod_example);


	/* Feature: unloading a library
	   Level:   recommended

		 The module gets uninitialized, memory freed.
	*/
	gpmi_mod_unload(mod_example);


	/* Feature: uninitialize gpmi
	   Level:   recommended

		 The library gets uninitialized, all memory are freed. You may
		 reinitialize it later.
	*/
	gpmi_uninit();

	printf("Example finished.\n");

	return 0;
}
