/**
 * @file events.c
 * \brief Handles the event-calls in GPMI.
 *
 * Handles creating, binding, calling and removing events within GPMI and handles
 *  all the calls to the modules and packages.
 */

#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <assert.h>
#include "events.h"
#include "error.h"
#include "threads.h"

/**
 * \brief Registers an event.
 *
 * @param event_name The name of the new event.
 * @param param_types The params the event should have.
 * @return The unique-ID for this event assigned by GPMI. You need to use this ID to call this event.
 */
int gpmi_event_register(const char *event_name, const char *param_types)
{
	int event_num;
	int old_len;

	/* Sanity check of parameters */
	assert(event_name != NULL && param_types != NULL);

	/* Acquire a lock for the global structure as we will need to modify the event database heavily */
	gpmi_shared_lock;

	/* Store the event name, auto increment, we get the unique event number */
	event_num = gpmi_keyword_add_auto(gpmi_shared.gpmi_events, event_name);

	/* gpmi_keyword_add_auto may fail */
	if (event_num == gpmi_err_kw_already_exists) {
		return gpmi_keyword_find_int(gpmi_shared.gpmi_events, event_name);
	}
	assert(event_num >= 0);

	/* Make more room in the types array if needed */
	if (event_num >= gpmi_shared.event_types_allocated) {
		old_len = gpmi_shared.event_types_allocated;
		gpmi_shared.event_types_allocated = event_num + 16;
		gpmi_shared.event_types = realloc(gpmi_shared.event_types, gpmi_shared.event_types_allocated * sizeof(char *));

		/* If realloc failed, we've lost all our events - do panic! */
		assert(gpmi_shared.event_types != NULL);

		/* Clean newly allocated memory */
		memset(gpmi_shared.event_types + old_len, 0, (gpmi_shared.event_types_allocated - old_len) * sizeof(char *));
	}

	/* Store parameter types */
	gpmi_shared.event_types[event_num] = strdup(param_types);

	gpmi_shared_unlock;
	return event_num;
}

/**
 * \brief Find an event by its name.
 *
 *  If the param_types is not NULL, the param_types of the event is written there.
 * @param event_name The event-name to search for.
 * @param param_types Will get the params this event has, if not NULL.
 * @return Returns the ID of the event, if found. -1 if it was not found.
 */
int gpmi_event_find(const char *event_name, char **param_types)
{
	int event_num;

	/* First check if the name exists at all in the event list */
	gpmi_shared_lock;
	event_num = gpmi_keyword_find_int(gpmi_shared.gpmi_events, event_name);	

	if (event_num < 0) {
		/* Event not found, report it */
		if (param_types != NULL)
			*param_types = NULL;
		gpmi_shared_unlock;
		return -1;
	}

	/* Return the parameter types, if needed */
	if (param_types != NULL)
		*param_types = gpmi_shared.event_types[event_num];

	gpmi_shared_unlock;

	/* Return the event num */
	return event_num;
}

/**
 * \brief Get the param_types of an event by its ID.
 *
 * @param event_num the ID of the event.
 * @return The param_types of the event.
 */
const char *gpmi_event_get_types(int event_num)
{
	assert(event_num >= 0 && event_num < gpmi_shared.event_types_allocated);

	/* Return the params */
	return gpmi_shared.event_types[event_num];
}

/**
 * \brief Initialize the event-system.
 *
 *   Is automaticly called internally.
 */
void gpmi_event_init()
{
	gpmi_shared_lock;

	/* Init events table */
	gpmi_shared.gpmi_events = gpmi_keyword_create("gpmi_events", gpmi_keyword_type_integer);

	gpmi_shared_unlock;
}

/**
 * \brief Uninitialize the event-system.
 *
 *   Is automaticly called internally.
 */
void gpmi_event_uninit()
{
	int n;
#ifdef gpmi_malloc_cache
	gpmi_private_globals_t *priv;
#endif

#ifdef gpmi_malloc_cache
	priv = gpmi_threads_get_private();

	/* Free global event malloc cache */
	free(priv->global_argv);
	priv->global_argv  = NULL;
	priv->global_argvn = 0;
#endif

	gpmi_shared_lock;

	/* Exit if events were not initialized */
	if (gpmi_shared.gpmi_events == NULL) {
		gpmi_shared_unlock;
		return;
	}
	
	/* Cycle trough all events and free their types */
	for (n = 0; n < gpmi_shared.gpmi_events->end; n++)
		free(gpmi_shared.event_types[n]);
	
	free(gpmi_shared.event_types);

	/* Destroy events table */
	gpmi_keyword_destroy(gpmi_shared.gpmi_events);		

	gpmi_shared_unlock;
}

/**
 * \brief Bind the global event-handler of a module to a pointer.
 *
 * @param mod The module where the global event-handler is bind to.
 * @param handler The proc which should be called in case of an event.
 * @return The old proc which was bind to this module (NULL if none).
 */
gpmi_global_event_t gpmi_event_bind_global(gpmi_module *mod, const void *handler)
{
	gpmi_global_event_t old_proc;

	/* Save old pointer - for hooking */
	old_proc = mod->proc_EventGlobal;

	/* Store handler in the global event handler of the module*/
	mod->proc_EventGlobal = (gpmi_global_event_t)handler;

	return old_proc;
}

/**
 * \brief Bind an event of a module to a pointer.
 *
 * @param mod The module for which the bind should apply.
 * @param event The event to bind the proc to.
 * @param handler The proc which should be called in case of the event.
 * @return The old proc which was bind to this event (NULL if none).
 */
gpmi_ProcEvent *gpmi_event_bind(gpmi_module *mod, int event, const void *handler)
{
	gpmi_ProcEvent *old_proc;
	int old_len;

	/* If we are running out of the preallocated space, allocate some more */
	if (event >= mod->event_allocated) {
		old_len = mod->event_allocated;
		mod->event_allocated = event + 16;
		mod->proc_EventArray = realloc(mod->proc_EventArray, mod->event_allocated * sizeof(gpmi_ProcEvent *));

		assert(mod->proc_EventArray != NULL);

		/* Clean newly allocated memory */
		memset(mod->proc_EventArray + old_len, 0, (mod->event_allocated - old_len) * sizeof(gpmi_ProcEvent *));

		/* We didn't have a procedure already bound */
		old_proc = NULL;
	} else {
		old_proc = mod->proc_EventArray[event];
	}

	/* Store the handler in the array */
	mod->proc_EventArray[event] = (gpmi_ProcEvent *)handler;
	if (event >= mod->event_next_free)
		mod->event_next_free = event + 1;

	return old_proc;
}

/**
 * \brief Close a module's event system (free up memory).
 */
void gpmi_event_close(gpmi_module *mod)
{
	/* Free the memory if allocated */
	if (mod->proc_EventArray != NULL)
		free(mod->proc_EventArray);

	mod->event_next_free  = 0;
	mod->event_allocated  = 0;
	mod->proc_EventGlobal = NULL;
	mod->proc_EventArray  = NULL;
}

/**
 * \brief Open a module's event system.
 */
void gpmi_event_open(gpmi_module *mod)
{
	mod->proc_EventGlobal = NULL;
	mod->proc_EventArray  = NULL;
	mod->event_next_free  = 0;
	mod->event_allocated  = 0;
}


/**
 * \brief internal: gpmi_event uses this vararg function to call a global event;
 * this is better to keep parsing vararg here than in the module because this way
 * global events are hookable.
 */
void gpmi_global_event_parse(gpmi_module *module, int event_id, ...)
{
	const char *types;
	gpmi_globev_param_t *argv, *arg;
	int argc;
	va_list ap;
	gpmi_global_event_t proc;
#ifdef gpmi_malloc_cache
	gpmi_private_globals_t *priv;
#endif

	proc = module->proc_EventGlobal;
	va_start(ap, event_id);
	argc = 0;

	/* First, count how many params we (should) have */
	types = gpmi_event_get_types(event_id);
	argc = strlen(types) + 1;

	/* Allocate the memory - may be cached */
#ifdef gpmi_malloc_cache
	priv = gpmi_threads_get_private();
	
	/* Enlarge the cache if needed */
	if (priv->global_argvn < argc) {
		priv->global_argvn = argc;	
		priv->global_argv = realloc(priv->global_argv, sizeof(gpmi_globev_param_t) * priv->global_argvn);
	}		
	argv = priv->global_argv;
#else
	/* Not cached: just malloc it */
	argv = malloc(sizeof(gpmi_globev_param_t) * argc);
#endif

	/* Read all params */
	for (arg = &(argv[0]), types = gpmi_event_get_types(event_id); (*types) != '\0'; types++, arg++) {
		switch (*types) {
			case 's':
			case 'S':
				arg->type   = gpmi_globev_type_string;
				arg->data.string = va_arg(ap, char *);
				break;
			case 'C':
				arg->type   = gpmi_globev_type_uchar;
				arg->data.integer = va_arg(ap, int);
				break;
			case 'b':
			case 'B':
			case 'i':
			case 'I':
				arg->type    = gpmi_globev_type_int;
				arg->data.integer = va_arg(ap, int);
				break;
			case 'k':
			case 'K':
				arg->type    = gpmi_globev_type_keyword;
				arg->data.integer = va_arg(ap, int);
				break;
			case 'u':
			case 'U':
				arg->type     = gpmi_globev_type_uint;
				arg->data.uinteger = va_arg(ap, unsigned int);
				break;
			/* XXX: we need better pointer and double handling! */
			/* Default should generate error */
			default:
				arg->type = gpmi_globev_type_ptr;
				arg->data.ptr  = va_arg(ap, void *);
				break;
		}
	}

	/* Done reading */
	va_end(ap);

	proc(module, event_id, argc, argv);

#ifndef gpmi_malloc_cache
	/* Free arguments if we are not caching */
	free(argv);
#endif
}

#ifdef NO_VARARG_MACRO
/* This means you can pass 200 params via this event.. in fact: it always
 *  puts 200 params over the stack. So lower this value if you want to save
 *  some stack-memory, but never go over this value with your params, else
 *  you just get random junk */
/* Also, make sure if you change this number, to change the macros of MSVC
 *  to match the COUNT here, else things can go badly wrong */
#define MAX_ARGS_COUNT 200
#define MAX_ARGS_SIZE (MAX_ARGS_COUNT * sizeof(void *))

void gpmi_event(gpmi_module *module, int event, ...)
{
#ifdef __GNUC__
	/* The GCC way.. easy and pretty fast */

	void *args;

	args = __builtin_apply_args();

#define call_func(func) __builtin_apply((void *)func, args, MAX_ARGS_SIZE);

#elif defined(_MSC_VER)
	/* The MSVC (and possible others) way.. ugly and slow */

	/* All these defines are here to create an arglist : args[0], args[1], ... , args[199]*/
#define arglistx10(args, i, j) \
	args[(i * 100) + (j * 10) + 0], \
	args[(i * 100) + (j * 10) + 1], \
	args[(i * 100) + (j * 10) + 2], \
	args[(i * 100) + (j * 10) + 3], \
	args[(i * 100) + (j * 10) + 4], \
	args[(i * 100) + (j * 10) + 5], \
	args[(i * 100) + (j * 10) + 6], \
	args[(i * 100) + (j * 10) + 7], \
	args[(i * 100) + (j * 10) + 8], \
	args[(i * 100) + (j * 10) + 9]

#define arglistx100(args, i) \
	arglistx10(args, i, 0), \
	arglistx10(args, i, 1), \
	arglistx10(args, i, 2), \
	arglistx10(args, i, 3), \
	arglistx10(args, i, 4), \
	arglistx10(args, i, 5), \
	arglistx10(args, i, 6), \
	arglistx10(args, i, 7), \
	arglistx10(args, i, 8), \
	arglistx10(args, i, 9)

#define arglist(args) \
	arglistx100(args, 0), \
	arglistx100(args, 1)

	va_list ap;
	void **args;
	int i;

	args = (void **)malloc(MAX_ARGS_SIZE);
	memset(args, 0, MAX_ARGS_SIZE);

	va_start(ap, event);
	for (i = 0; i < MAX_ARGS_COUNT; i++)
		args[i] = va_arg(ap, void *);
	va_end(ap);

#define call_func(func) func(module, event, arglist(args))

#endif

#ifndef call_func
#error We are sorry, but we don't have any way to handle variadic macros for this compiler. Feel free to supply us with one. Can't continue, aborting..
#else

	/* The real function */
	if (event >= 0 && module->proc_EventGlobal != NULL) {
		gpmi_set_current_module(module);
		call_func(gpmi_global_event_parse);
	}
	if (event >= 0 && event < module->event_next_free) {
		if (module->proc_EventArray[event] != NULL) {
			gpmi_set_current_module(module);
			call_func(module->proc_EventArray[event]);
		}
	}

#undef call_func
#endif
}
#endif
