/*
    hash 1.0.0 simple hash tables with linked lists
    Copyright (C) 2004 Tibor `Igor2` Palinkas

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Author: igor2gpmi@peticio.hu
*/



#include <stdlib.h>
#include <stdio.h> //only for stderr, will be removed later
#include "hash.h"
#include <string.h>
#include <ctype.h>

//store a data-key pair
void hash_store(struct hash_t *h, const void *key, const void *data)
{
	struct hdata_t *d;
	struct hdata_t **a, *b;
	unsigned int (*hash_calc) (int cmd, const void *key, const void *key2, char **result_pointer);
	int h_;

	hash_calc = h->hash_calc;


	/* allocate and fill a new data structure */
	d = malloc(sizeof(struct hdata_t));
	d->next = NULL;

	if (hash_calc(h2_cmd_allockey, key, NULL, (char **)&(d->key)) != 0)
		fprintf(stderr, "Hash2: probably bad defcalc (old API?), returns non-zero for allockey!\n");

	d->data = (void *)data;

	/* calculate a hash value for the key */
	h_ = (hash_calc(h2_cmd_key, d->key, NULL, NULL)) % h->tbl_size;

	/* let's see if anything is stored under that key */
	a = ((h->tbl) + h_);

	/* no, it was empty, we creat a record */
	if (*a == NULL)
		*a = d;
	else {
		/* have to seek the end of the linked list at the given tbl pos */
		b = *a;
		do {
			if (b->next == NULL) {
				/* last record */
				b->next = d;
				return;
			}
			b = b->next;
		} while (1);
	}
	/* we shouldn't ever get here... */
}

/* find a key */
struct hdata_t *hash_find_(struct hash_t *h, const void *key)
{
	int h_;
	struct hdata_t **a, *b;
	unsigned int (*hash_calc) (int cmd, const void *key, const void *key2, char **result_pointer);

	if (h->tbl_size <= 0)
		return NULL;

	/* calculate a hash value for the key, fetch the given node */
	hash_calc = h->hash_calc;
	h_ = (hash_calc(h2_cmd_key, key, NULL, NULL)) % h->tbl_size;
	a = (h->tbl) + h_;

	/* If the table entry is empty, sure we don't know this key */
	if (*a == NULL)
		return NULL;

	/* the table entry contains at least one record, linear search in a linked list */
	b = *a;
	do {
		if ((b->key != NULL) && (hash_calc(h2_cmd_cmpkey, key, b->key, NULL) == 0))
			return b;
		b = b->next;
	} while (b != NULL);

	return NULL;
}


const void *hash_find(struct hash_t *h, const void *key)
{
	struct hdata_t *a;

	a = hash_find_(h, key);
	return (a == NULL) ? NULL : a->data;
}

struct hash_t *hash_create(int tbl_size, void *calc)
{
	struct hash_t *h;

	h = malloc(sizeof(struct hash_t));
	h->tbl_size = tbl_size;

	if (calc == NULL)
		h->hash_calc = hash_defcalc;
	else
		h->hash_calc = calc;

	h->tbl = calloc(sizeof(struct hdata_t), tbl_size);
	return h;
}

struct hash_t *hash_rehash(struct hash_t *h, int new_size, void *new_calc)
{
	struct hash_t *h2;
	struct hdata_t **a, *b, *c;
	int n;

	if (new_calc != NULL)
		h2 = hash_create(new_size, new_calc);
	else
		h2 = hash_create(new_size, h->hash_calc);

	for (a = h->tbl, n = 0; n < h->tbl_size; n++, a++) {
		b = *a;
		if (b != NULL) {
			do {
				hash_store(h2, b->key, b->data);
				c = b->next;
				free(b->key);
				free(b);
				b = c;
			} while (b != NULL);
		}
	}
	free(h->tbl);
	h->tbl = NULL;
	h->tbl_size = 0;
	free(h);
	return h2;
}


int hash_destroy(struct hash_t *h)
{
	unsigned int (*hash_calc) (int cmd, void *key, void *key2, char **result_pointer);
	struct hdata_t **a, *b, *c;
	int n;
	hash_calc = h->hash_calc;

	for (a = h->tbl, n = 0; n < h->tbl_size; n++, a++) {
		b = *a;
		if (b != NULL) {
			do {
				c = b->next;
				hash_calc(h2_cmd_freekey, b->key, NULL, NULL);
				hash_calc(h2_cmd_destroy, b, NULL, NULL);				
				free(b);
				b = c;
			} while (b != NULL);
		}
	}
	free(h->tbl);
	h->tbl = NULL;
	h->tbl_size = 0;
	free(h);
	return 0;
}

unsigned int hash_defcalc(int cmd, char *str, char *str2, char **result_pointer)
{
	char *s;
	unsigned char c;
	unsigned int a, n;

	switch (cmd) {

	case h2_cmd_key:
		for (a = 0, n = 0, s = str; (*s != 0) && (*s != ' ') && (*s != '\t') && (*s != '\n') && (*s != '\r'); s++, n++) {
			c = toupper(*s);
			a += (((c - 48) % 128) << (n));
		}
		return a;
		break;

	case h2_cmd_cmpkey:
		if (strcasecmp(str, str2) == 0)
			return 0;
		return 1;

	case h2_cmd_cmprec:
		if (strcasecmp(str, str2) == 0)
			return 0;
		return 1;

	case h2_cmd_allockey:
		hash2_astrcpy(s, str);
		*result_pointer = s;
		return 0;

	case h2_cmd_freekey:
		free(str);
		return 0;

	case h2_cmd_destroy:
		return 0;
	}
	return -1;
}

int hash_del_key(struct hash_t *h, void *key)
{
	int h_;
	struct hdata_t **a, *b, *bl;
	unsigned int (*hash_calc) (int cmd, const void *key, const void *key2, char **result_pointer);


	//calculate a hash value for the key
	hash_calc = h->hash_calc;
	h_ = (hash_calc(h2_cmd_key, key, NULL, NULL)) % h->tbl_size;
	a = (h->tbl) + h_;

	/* If the table entry is empty, sure we don't know this key */
	if (*a == NULL)
		return 0;

	//the table entry contains at least one record, linear search in a linked list
	b = *a;
	bl = NULL;
	do {
		if ((b->key != NULL) && (hash_calc(h2_cmd_cmpkey, key, b->key, NULL) == 0)) {

			if (bl == NULL)
				*a = NULL;
			else
				bl->next = b->next;

			hash_calc(h2_cmd_freekey, b->key, NULL, NULL);
			hash_calc(h2_cmd_destroy, b, NULL, NULL);
			free(b);
			return 1;
		}
		bl = b;
		b = b->next;
	} while (b != NULL);
	return 0;
}


int hash_del_rec(struct hash_t *h, const void *key, const void *data)
{
	int h_;
	struct hdata_t **a, *b, *bl;
	unsigned int (*hash_calc) (int cmd, const void *key, const void *key2, char **result_pointer);


	//calculate a hash value for the key
	hash_calc = h->hash_calc;
	h_ = (hash_calc(h2_cmd_key, key, NULL, NULL)) % h->tbl_size;
	a = (h->tbl) + h_;

	/* If the table entry is empty, sure we don't know this key */
	if (*a == NULL)
		return 0;

	/* the table entry contains at least one record, linear search in a linked list */
	b = *a;
	bl = NULL;

	do {
		if ((b->key != NULL)
				&& (hash_calc(h2_cmd_cmpkey, key, b->key, NULL)   == 0)
				&& (hash_calc(h2_cmd_cmprec, data, b->data, NULL) == 0)) {

			if (bl == NULL)
				*a = b->next;
			else
				bl->next = b->next;

			hash_calc(h2_cmd_freekey, b->key, NULL, NULL);
			hash_calc(h2_cmd_destroy, b, NULL, NULL);			
			free(b);
			return 1;
		}
		bl = b;
		b = b->next;
	} while (b != NULL);
	return 0;
}
