#include <stdlib.h>
#include <stdio.h>
#include "libs.h"
#include "os_dep.h"
#include "error.h"

#ifdef WINDLL
#include <assert.h>
char *gpmi_dll_dir = "gpmi_dll";
#endif

#ifdef WIN32
#include <windows.h>
#include <string.h> // for strrchr()

/* Function basename is not native on windows */
char *basename(char *path)
{
	char* last = strrchr(path, '/');
	if ((path == NULL) || (*path == '\0')) return ".";
	if (*path == '/' && *(path + 1) == '\0') return "/";
	if (last == NULL) return path;
	if (last && *(last + 1) == '\0') {
		*last = '\0';
		last = strrchr(path, '/');
	}
	if (last) return (++last);
	return path;
}

/* Function dirname is not native on windows */
char *dirname(char *path)
{
	char* last = strrchr(path, '/');
	if ((path == NULL) || (*path == '\0')) return ".";
	if (*path == '/' && *(path + 1) == '\0') return "/";
	if (last == NULL) return ".";
	if (last && *(last + 1) == '\0') {
		*last = '\0';
		last = strrchr(path, '/');
	}
	if (last) *last = '\0';
	if (*path == '\0') return "/";
	return path;
}

#endif

/**
 * Build full pathname of a library and its dep
 */
void gpmi_library_add_path(const char *dir, const char *library_name, char *lib_file_name, char *dep_file_name, int path_size)
{

/* Os dependent part: the file name of a dynamic lib */
#ifdef WINDLL
		char *sdir, *s, *s2, last;

		/* This is a bit ugly. We need to convert paths into filenames and
		   collect all dlls in a single dir. However, mostly because the dep
			 file, we need to insert the path of the dll collection */

		/* not to destroy the dir string */
		sdir = strdup(dir);

		/* all '/' and '\' is changed to '_'; replace ../ with @@@ */
		for(s = sdir; *s != '\0'; s++) {
			switch(*s) {
				case '.':
					if ((*(s+1) == '.') && ((*(s+2) == '\\') || (*(s+2) == '/'))) {
						*s = '@';
						*(s+1) = '@';
						*(s+2) = '@';

						s++;

					}
					break;
				case '\\':
				case '/':
					*s = '_';
					break;
			}
		}

		/* Eat up @ characters */
		last = '\0';
		for(s = s2 = sdir; *s != '\0'; s++) {
			switch(*s) {
				case '@':
					break;
				case '\\':
					if (last == '\\')
						break;
				default:
					last = *s2 = *s;
					s2++;
			}
		}
		*s2 = '\0';

		/* Now assamble the paths */
		snprintf(lib_file_name,	path_size,	"%s\\%s_%s.dll", gpmi_dll_dir, sdir, library_name);
		snprintf(dep_file_name,	path_size, "%s\\%s_%s.dep", gpmi_dll_dir, sdir, library_name);

/*		printf("lib_file_name = %s\n", lib_file_name);
		printf("dep_file_name = %s\n", dep_file_name);*/

		/* free temporary storage */
		free(sdir);

#elif DLCOMPAT
		/* With dlcompat, just add dir, file name and extension */
		snprintf(lib_file_name,	path_size,	"%s/%s.dlc",	dir, library_name);
		snprintf(dep_file_name,	path_size,	"%s/%s.dep",	dir, library_name);
#elif DARWIN
		/* On darwin, it is like unix, only so -> dylib */
		snprintf(lib_file_name, path_size,      "%s/%s.dylib",     dir, library_name);
		snprintf(dep_file_name, path_size,      "%s/%s.dep",    dir, library_name);
#else
		/* On unix, just add dir, file name and extension */
		snprintf(lib_file_name,	path_size,	"%s/%s.so",		dir, library_name);
		snprintf(dep_file_name,	path_size,	"%s/%s.dep",	dir, library_name);
#endif
}

/**
 *  Os dependent part for loading a dynamic linked library
 */
int gpmi_library_load_lowlevel(gpmi_library *library, char *file_name)
{
	char *error_str;
#ifdef WINDLL
/* XXX - this is another glitch. SetDllDirectory would be the proper way to
   tell windows to look for gpmi packages/modules in our gpmi dir. But of
	 course, we can not use it. It is supported only by higher versions of
	 windows (from 0x0502 which means anything upper than windows XP that
	 is only windows server 2003 and longhorn, it seems).

		SetDllDirectory(gpmi_dll_dir);

   Instead, we will change current directory before loading, and change it
	 back after, which would work from 95 up to longhorn. Windows API reference
	 does not recommend doing this, but using SetDllDirectory.

	 What else could I say... You wanted windows, now enjoy it! :)
*/
	/* Workaround instead of SetDllDirectory starts here */
	char cwd[MAX_PATH+2];

	GetCurrentDirectory(MAX_PATH+1, cwd);
	SetCurrentDirectory(gpmi_dll_dir);
	/* Workaround instead of SetDllDirectory ends here*/

//	library->handle = LoadLibraryEx(file_name, NULL, 0);
	library->handle = LoadLibrary(file_name);

	if (!library->handle) {
		error_str = malloc(256);
		snprintf(error_str, 256, "code %d on file %s", GetLastError(), file_name);
		gpmi_err_stack_push(gpmi_err_load_library, gpmi_err_src_lib_load_lowlevel, error_str);
		free(error_str);
		return 1;
	}
	/* Workaround instead of SetDllDirectory starts here */
	SetCurrentDirectory(cwd);
	/* Workaround instead of SetDllDirectory ends here */
#else
	library->handle = dlopen(file_name, RTLD_GLOBAL | RTLD_LAZY);
	if (library->handle == NULL) {
		error_str = malloc(256);
		snprintf(error_str, 256, "%s on file %s", dlerror(), file_name);
		gpmi_err_stack_push(gpmi_err_load_library, gpmi_err_src_lib_load_lowlevel, error_str);
		free(error_str);
		return 1;
	}
#endif

	return 0;
}

/**
 *  OS dependent part to unload a dynamic link library
 */
void gpmi_library_unload_lowlevel(gpmi_library *library)
{
#ifdef WINDLL
		FreeLibrary(library->handle);
#else
		dlclose(library->handle);
#endif
}

/**
 *  OS dependent part to resolve a symbol
 */
void *gpmi_dlsym_lowlevel(gpmi_library *library, const char *name)
{
#ifdef WINDLL
		return GetProcAddress(library->handle, name);
#else
		return dlsym(library->handle, name);
#endif
}

/**
 *  OS dependent part to get a handle to the global address space
 */
void *gpmi_dlopen_global()
{
#ifdef WINDLL
		assert(!"Your program has static linked GPMI packages. Unfortunately DLLs can't provide the functionality needed for this, you need to use dlc or use dynamic packages only.");
#else
		return dlopen(NULL, RTLD_GLOBAL);
#endif
}

int gpmi_install_root(char *inst_root, int inst_root_size)
{
#ifdef WIN32
	HKEY hKey;
	LONG lRet;
	char buf[PATH_MAX];
	DWORD size;

	// check HKCU\SOFTWARE\LibGPMI\InstallPath
	lRet = RegOpenKeyEx(HKEY_CURRENT_USER, "SOFTWARE\\LibGPMI", 0, KEY_QUERY_VALUE, &hKey);
	if (lRet == ERROR_SUCCESS) {
		size = inst_root_size;
		lRet = RegQueryValueEx(hKey, "InstallPath", NULL, NULL, (LPBYTE)buf, &size);
		RegCloseKey(hKey);
		if (lRet == ERROR_SUCCESS) {
			snprintf(inst_root, inst_root_size, buf);
			return 0;
		}
	}

	// check HKLM\SOFTWARE\LibGPMI\InstallPath
	lRet = RegOpenKeyEx(HKEY_LOCAL_MACHINE, "SOFTWARE\\LibGPMI", 0, KEY_QUERY_VALUE, &hKey);
	if (lRet == ERROR_SUCCESS) {
		size = inst_root_size;
		lRet = RegQueryValueEx(hKey, "InstallPath", NULL, NULL, (LPBYTE)buf, &size);
		RegCloseKey(hKey);
		if (lRet == ERROR_SUCCESS) {
			snprintf(inst_root, inst_root_size, buf);
			return 0;
		}
	}
#endif
	snprintf(inst_root, inst_root_size, GPMI_INST_ROOT);
	return 0;
}

#ifdef WIN32
char *gpmi_init_path(char *path)
{
	char inst_root[PATH_MAX];
	char buf[PATH_MAX];

	gpmi_install_root(inst_root, sizeof(inst_root));
	snprintf(buf, sizeof(buf), "%s%s", inst_root, path);
	return strdup(buf);	
}
#endif
