#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "paths.h"
#include "threads.h"

#define path_fetch(dest, path) \
	switch(path) { \
		case GPMI_PATH_SCRIPTS:  dest = &(gpmi_shared.path_scripts);  break; \
		case GPMI_PATH_MODULES:  dest = &(gpmi_shared.path_modules);  break; \
		case GPMI_PATH_PACKAGES: dest = &(gpmi_shared.path_packages); break;	\
		default: \
			assert(!"Unknown path id"); \
			exit(195); \
	}

void gpmi_path_append(gpmi_paths_t gpmi_path, const char *path)
{
	char **dest;

	path_fetch(dest, gpmi_path);

	if (gpmi_initialized == 0) {
		/* If we are not initialized, we should just put the path behind the pointer */
		gpmi_threads_init();
		*dest = strdup(path);
		return;
	}

	gpmi_shared_lock;
	gpmi_path_append_(dest, path);
	gpmi_shared_unlock;
}

void gpmi_path_insert(gpmi_paths_t gpmi_path, const char *path)
{
	char **dest;

	path_fetch(dest, gpmi_path);

	if (gpmi_initialized == 0) {
		/* If we are not initialized, we should just put the path behind the pointer */
		gpmi_threads_init();
		*dest = strdup(path);
		return;
	}

	gpmi_shared_lock;
	gpmi_path_insert_(dest, path);
	gpmi_shared_unlock;
}

int gpmi_path_find(gpmi_paths_t gpmi_path, const char *path)
{
	char **dest;
	int ret;

	if (gpmi_initialized == 0) {
		/* If we are not initialized, sure we don't have the path */
		return -1;
	}

	path_fetch(dest, gpmi_path);

	gpmi_shared_lock;
	ret = gpmi_path_find_((const char **)dest, path);
	gpmi_shared_unlock;

	return ret;
}

int gpmi_path_delete(gpmi_paths_t gpmi_path, const char *path)
{
	char **dest;
	int ret;

	if (gpmi_initialized == 0) {
		/* If we are not initialized, sure we don't have the path */
		return -1;
	}

	path_fetch(dest, gpmi_path);

	gpmi_shared_lock;
	ret = gpmi_path_delete_(dest, path);
	gpmi_shared_unlock;

	return ret;
}

/* --------------------------- String handlers ---------------------------- */

void gpmi_path_append_(char **gpmi_path, const char *path)
{
	char *tmp;
	int len;

	if (*gpmi_path != NULL) {
		/* Append at the end */
		len = strlen(*gpmi_path);
		tmp = malloc(len + strlen(path) + 2);

		/* Assamble the string; I avoid strcat here to save some cpu */
		{
			/* First copy the original path list - known length */
			memcpy(tmp, *gpmi_path, len);

			/* Append a ':' (overwrite a character after the path is faster) */
			tmp[len] = GPMI_PATH_SEP_CHAR;

			/* Copy the new path after the ':' */
			strcpy(tmp+len+1, path);
		}

		free(*gpmi_path);
		*gpmi_path = tmp;
	}
	else {
		/* Empty destination, just replace it */
		*gpmi_path = strdup(path);
	}
}


void gpmi_path_insert_(char **gpmi_path, const char *path)
{
	char *tmp;
	int len;

	if (*gpmi_path != NULL) {
		/* Insert at the beginning */
		len = strlen(path);
		tmp = malloc(len + strlen(*gpmi_path) + 2);

		/* Assamble the string; I avoid strcat here to save some cpu */
		{
			/* First copy the new path - known length */
			memcpy(tmp, path, len);

			/* Append a ':' (overwrite a character after the path is faster) */
			tmp[len] = GPMI_PATH_SEP_CHAR;

			/* Copy the original path list after the ':' */
			strcpy(tmp+len+1, *gpmi_path);
		}

		free(*gpmi_path);
		*gpmi_path = tmp;
	}
	else {
		/* Empty destination, just replace it */
		*gpmi_path = strdup(path);
	}
}


/* Returns the character position of the match or -1 if not found */
int gpmi_path_find_(const char **gpmi_path, const char *path)
{
	const char *s, *ss;
	int n, start;

	if (*gpmi_path == NULL)
		return -2;

	for(s=*gpmi_path, ss=path, n=0, start=0; ; s++, n++) {
		switch (*s) {
			case GPMI_PATH_SEP_CHAR:
			case '\0':
				/* Found a separator, check if search string is over */
				if ((ss != NULL) && (*ss == '\0')) {
					/* It is over, so we have a match. Return where it began */
					return start;
				}
				else {
					/* Test string is not over; is path string over? */
					if (*s == '\0')
						return -1;

					/* Reset start and the search string so we cna restart at the next token */
					start = n + 1;
					ss = path;
				}
				break;
			default:
				/* Found a normal character, compare if test string was ok so far */
				if (ss != NULL)
					if (*ss != *s)
						ss = NULL;
		}
	}

	/* We shouldn't ever get here */
	return -3;
}

/* Returns 0 on success and negative value on error */
int gpmi_path_delete_(char **gpmi_path, const char *path)
{
	int pos, len;
	char *next;

	/* Search the path. */
	pos = gpmi_path_find_((const char **)gpmi_path, path);
	if (pos < 0)
		return pos;

	len  = strlen(*gpmi_path);
	next = strchr(*gpmi_path + pos, GPMI_PATH_SEP_CHAR);

	if (next != NULL) {
		/* shift the rest of the string back a bit */
		strcpy(*gpmi_path + pos, next);
	}
	else {
		/* Just truncate the string */
		*(*gpmi_path + pos - 1) = '\0';
	}

	return 0;
}
