#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <assert.h>
#include "libs.h"
#include "libs-package.h"
#include "global.h"
#include "pkgs-static.h"
#include "os_dep.h"
#include "threads.h"

/**
 * Lowlevel function to "load" a static library
 */
static gpmi_library *gpmi_load_library_static(const char *library_name)
{
	char file_name[PATH_MAX];
	gpmi_library *library;

	snprintf(file_name, sizeof(file_name), "__STATIC__%s", library_name);

/* Paranoid feature: check for double linked static libs */
	if (gpmi_find_library(file_name) != NULL) {
		fprintf(stderr, "GPMI: static package %s is linked in multiple times.\n", file_name);
		assert(!"Multiple static linking of the same package");
	}

	library = gpmi_lib_create(file_name);
	library->handle     = gpmi_dlopen_global();
	library->flags      = gpmi_libf_static;
	library->references = 1;

	return library;
}

/**
 * Is called when a static package is found at startup (__construct__).
 *  We make a package available, and make sure we can find it back.
 */
void gpmi_static_pkg_register(char *name, void *pkg_init, void *pkg_uninit,  void *pkg_checkver,  void *pkg_register_commands)
{
	gpmi_package *package;

	/* Init GPMI */
	gpmi_initialize_paths = 0;
	gpmi_init();

	/* Create the package */
	package = malloc(sizeof(gpmi_package));
	memset(package, 0, sizeof(package));

	/* Find the library */
	package->library = gpmi_load_library_static(name);

	/* Set the procs */
	package->proc_Init     = pkg_init;
	package->proc_UnInit   = pkg_uninit;
	package->proc_CheckVer = pkg_checkver;
	package->proc_RegCmd   = pkg_register_commands;

	/* Make sure we can use this multiple times from library level */
	package->library->static_package = package;

	/* Add the package to the static_package list */
	gpmi_shared_lock;
	package->next = gpmi_shared.static_packages;
	gpmi_shared.static_packages = package;

	/* Init the package */
	package->proc_Init();
	gpmi_shared_unlock;
}

/**
 * Is called when a static package is stopped (__destructor__).
 *  Here we remove it from the linked list, and uninitialize GPMI if needed.
 */
void gpmi_static_pkg_unregister(char *name)
{
	gpmi_package *pkg, *last;
	char library_name[PATH_MAX];

	gpmi_shared_lock;
	
	/* Start at the beginning */
	pkg  = gpmi_shared.static_packages;
	last = NULL;

	snprintf(library_name, sizeof(library_name), "__STATIC__%s", name);

	/* Find our package */
	while (pkg != NULL) {
		assert(pkg->library->name != NULL);
		if (strcmp(pkg->library->name, library_name) == 0) {
			/* This is our package */

			/* Remove from list */
			if (last == NULL)
				gpmi_shared.static_packages = pkg->next;
			else
				last->next = pkg->next;

			/* Unload package */
			gpmi_pkg_unload(pkg);
			gpmi_pkg_free(pkg);
			gpmi_shared_unlock;

			/* In case we are the last package, unload GPMI */
			gpmi_uninit();
			return;
		}

		last = pkg;
		pkg = pkg->next;
	}

	/* It should never come here.. if it does, something went really wrong:
	   the package tries to unregister is already unregistered! */
	assert(0);
}
