/* STUTTER GPMI module */

#include "stutter.h"

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>

#include "gpmi/debug.h"


/**
 * Returns the extension of this script-module.
 */
char *Stutter_GetExtension()
{
	return strdup(".stt");
}

/**
 * Reloads the script on the fly.
 */
int Stutter_Rehash(void *client_data)
{
	Stutter_ClientData *MyData = client_data;
	char *name;
	int r;

	name = MyData->name;
	if (!name)
		return 0;

	r = Stutter_LoadScript(client_data, name);
	free(name);
	return r;
}

/**
 * Loads a script.
 */
int Stutter_LoadScript(void *client_data, const char *filename)
{
	Stutter_ClientData *MyData = client_data;
	s_obj *res;
	s_mark mark;

	MyData->name = strdup(filename);

	mark = gc_prot_mark();

	res = evalfile(MyData->global, filename);
	if (res->type == ST_ERROR) {
		printobj(stderr, res);
		fprintf(stderr, "\n");
		gc_prot_free(mark, NULL);
		return 1;
	}

	gc_prot_free(mark, NULL);

	return 0;
}

/**
 * Execute a dynamic command.
 **/
s_obj *Stutter_DynamicPackageCommand(varctx *ctx, s_obj *parm, void *data, const int tail)
{
//	Stutter_ClientData *MyData = GPMI_Script_GetClientData();
	int argc;
	char **args;
	s_obj *obj;
	char *s;

	char * (* proc) (int, char **);

	proc = GPMI_Script_FindProc((char *) data);

	/* Check if command not found - this must be an internal error: the
	   package had registered the command in the hash, but the routine
	   pointer is NULL.
	*/
	if (proc == NULL) {
		gpmi_debug(0, "[Stutter] Internal error: %s has a NULL proc in the command hash\n", (char *) data);
		return create_nil();
	}

	argc = 0;
	for(obj = parm; obj->type == ST_CONS; obj = obj->d.cons.cdr) {
		argc ++;
	}

	args = (char **) malloc(sizeof(char *) * (argc+1));
	assert(args != NULL);

	argc = 1;
	while(nextarg(ctx, &obj, &parm)) {
		if(obj->type != ST_STRING) {
			fprintf(stderr, "DynamicPackageCommand fails with SE_TYPE\n");
			return create_error(SE_TYPE, "Type mismatch");
		}
		args[argc++] = obj->d.string.value;
	}
	args[0] = (char *) data;

	s = proc(argc, args);
	free(args);

	if (!s)
		return create_nil();

	return create_string(s);
}

/**
 * Register a command in our command hash
 */
void Stutter_RegisterCommand(void *client_data, const char *name)
{
	Stutter_ClientData *MyData = client_data;
	char *str, *sym, *c;
	s_mark mark;
	Stutter_FuncNames *fn;

	str = strdup(name);
	sym = strdup(name);
	for(c = sym; *c; c++) *c = toupper(*c);

	mark = gc_prot_mark();	

	varctx_set(MyData->global, sym,
			create_builtin_with_data(Stutter_DynamicPackageCommand, str));
	free(sym);

	fn = (Stutter_FuncNames *) (malloc(sizeof(Stutter_FuncNames)));
	assert(fn != NULL);
	fn->name = str;
	fn->next = MyData->names;
	MyData->names = fn;

	gc_prot_free(mark, NULL);
}

/**
 * Call a function of the script by name, passing params.
 */
void Stutter_RunFunction(void *client_data, const char *function, int argc, const char **argv)
{
	Stutter_ClientData *MyData = client_data;
	s_obj *call;
	s_obj *tail;
	int i;
	s_mark mark;

	mark = gc_prot_mark();

	call = create_cons(create_symbol(function), NULL);
	tail = call;

	for(i=0;i<argc;i++) {
		tail->d.cons.cdr = create_cons(create_string(argv[i]), NULL);
		tail = tail->d.cons.cdr;
	}

	tail->d.cons.cdr = create_nil();

	call = eval(MyData->global, call);
	if (call->type == ST_ERROR) {
		printobj(stderr, call);
		fprintf(stderr, "\n");
	}

	gc_prot_free(mark, NULL);
}

int Stutter_Initialize(void **client_data)
{
	Stutter_ClientData *MyData;

	stutter_init();

	MyData = malloc(sizeof(Stutter_ClientData));
	memset(MyData, 0, sizeof(Stutter_ClientData));

	MyData->global = varctx_create(NULL, 1024);

	gc_protect_root(MyData->global);

	register_builtins(MyData->global);

//	GC_VERBOSE=1;
//	GC_ALLOC_LIMIT=128;

	*client_data = MyData;

	return 0;
}

void Stutter_Destroy(void **client_data)
{
	Stutter_ClientData *MyData = *client_data;
	Stutter_FuncNames *n, *m;

	if (!MyData->name)
		return; /* not initialized */

	if (MyData->name)
		free(MyData->name);


	varctx_free(MyData->global);
	gc_sweep(NULL);

	for(n=MyData->names; n; ) {
		m = n;
		n = n->next;
		free(m->name);
		free(m);
	}

	/* Finally free up the client-data */
	free(*client_data);

	stutter_shutdown();
}
