/**
 * Regression tests for keywords
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "gpmi.h"
#include "keywords.h"
#include "threads.h"


int verbose = 0;

#define Dprintf(x...)		if (verbose > 0) printf(x);

const gpmi_keyword_static_int array_int[] = {
	{"Sun", 1},
	{"Mon", 2},
	{"Tue", 3},
	{"Wed", 4},
	{"Thu", 5},
	{"Fri", 6},
	{"Sat", 7},
	{NULL, 0}
} ;

const gpmi_keyword_static_int array_bits[] = {
	{"meat", 		1},
	{"onion", 	2},
	{"potato",	4},
	{"rice", 		8},
	{"spices", 	16},
	{NULL, 0}
} ;

const gpmi_keyword_static_string array_string[] = {
	{"bbl", "be back later"},
	{"brb", "be right back"},
	{"gtg", "got to go"},
	{"iwfg", "it works for good"},
	{"lol", "laughing out loud"},
	{"rotf", "rolling on the floor"},
	{NULL, 0}
} ;

gpmi_keyword *days, *food, *acrynoms;

void test_int() {
	int day;
	char *name;

	Dprintf("\n--- Testing integers \n");
	days = gpmi_keyword_import("Days", gpmi_keyword_type_integer, array_int);

	/* Add a new keyword */
	if (gpmi_keyword_add(days, "Holiday", 11) != 0)
		fprintf(stderr, "TEST FAILS: gpmi_keyword_find_int/1");

	/* Check enforcement of single keys */
	if (gpmi_keyword_add(days, "Holiday", 11) == 0)
		fprintf(stderr, "TEST FAILS: gpmi_keyword_find_int/2");

	/* Checking imported data */
	day = gpmi_keyword_find_int(days, "Fri");
	Dprintf("Fri=%d\n", day);
	if (day!=6)
		fprintf(stderr, "TEST FAILS: gpmi_keyword_find_int/3");

	name = (char *)gpmi_keyword_find_rev(days, 2);
	Dprintf("2=%s\n", name);
	if (strcmp(name, "Mon") != 0)
		fprintf(stderr, "TEST FAILS: gpmi_keyword_find_int/4");

	/* Checking dynamic added data */
	day = gpmi_keyword_find_int(days, "Holiday");
	Dprintf("Holiday=%d\n", day);
	if (day!=11)
		fprintf(stderr, "TEST FAILS: gpmi_keyword_find_int/5");

	name = (char *)gpmi_keyword_find_rev(days, 11);
	Dprintf("11=%s\n", name);
	if (strcmp(name, "Holiday") != 0)
		fprintf(stderr, "TEST FAILS: gpmi_keyword_find_int/6");


	/* Checking for invalid keywords */
	day = gpmi_keyword_find_int(days, "NoDay");
	Dprintf("NoDay=%d\n", day);
	if (day>0)
		fprintf(stderr, "TEST FAILS: gpmi_keyword_find_int/7");

	/* Checking for delete */
	gpmi_keyword_del(days, "Fri");
	day = gpmi_keyword_find_int(days, "Fri");
	Dprintf("Fri=%d\n", day);
	if (day==6)
		fprintf(stderr, "TEST FAILS: gpmi_keyword_del/8");

	name = (char *)gpmi_keyword_find_rev(days, 6);
	Dprintf("6=%s\n", name);
	if (name!=NULL)
		fprintf(stderr, "TEST FAILS: gpmi_keyword_del/9");
}

void test_multi() {
	int i;
	char *names;

	Dprintf("\n--- Testing multi \n");
	food = gpmi_keyword_import("food", gpmi_keyword_type_integer, array_bits);

	/* Check multiple keyword resolving */
	if (gpmi_keyword_find_multi(food, "meat+rice", &i)!=0)
		fprintf(stderr, "TEST FAILS: gpmi_keyword_find_multi/0");

	Dprintf("Multi: meat+rice=%d\n", i);

	if (i != 9) 	
		fprintf(stderr, "TEST FAILS: gpmi_keyword_find_multi/1");

	names = (char *)gpmi_keyword_find_rev_multi(food, 9);
	Dprintf("9=%s\n",names);
	if (strcmp("meat+rice", names) != 0)
		fprintf(stderr, "TEST FAILS: gpmi_keyword_find_multi/2");


	free(names);
}

void test_str() {
	char *text;
	char *name;

	Dprintf("\n--- Testing strings \n");

	acrynoms = gpmi_keyword_import("Acrynoms", gpmi_keyword_type_string, array_string);

	/* Add a new keyword */
	gpmi_keyword_add(acrynoms, "dak", strdup("dynamic added keyword"));


	/* Checking imported data */
	text = (char *)gpmi_keyword_find_string(acrynoms, "iwfg");
	Dprintf("iwfg=%s\n", text);
	if (strcmp(text, "it works for good")!=0)
		fprintf(stderr, "TEST FAILS: gpmi_keyword_find_str/1");

	name = (char *)gpmi_keyword_find_rev(acrynoms, "be right back");
	Dprintf("be right back=%s\n", name);
	if (strcmp(name, "brb") != 0)
		fprintf(stderr, "TEST FAILS: gpmi_keyword_find_str/2");

	/* Checking dynamic added data */
	text = (char *)gpmi_keyword_find_string(acrynoms, "dak");
	Dprintf("dak=%s\n", text);
	if (strcmp(text, "dynamic added keyword")!=0)
		fprintf(stderr, "TEST FAILS: gpmi_keyword_find_str/3");

	name = (char *)gpmi_keyword_find_rev(acrynoms, "dynamic added keyword");
	Dprintf("dynamic added keyword=%s\n", name);
	if (strcmp(name, "dak") != 0)
		fprintf(stderr, "TEST FAILS: gpmi_keyword_find_str/4");

}

void test_auto() {
	gpmi_keyword *kw;
	int id;
	char *str;

	Dprintf("\n--- Testing auto-increment list \n");
	kw = gpmi_keyword_create("Names", gpmi_keyword_type_integer);

	/* Add a new keyword */
	gpmi_keyword_add_auto(kw, "Susan");
	gpmi_keyword_add_auto(kw, "John");
	gpmi_keyword_add_auto(kw, "Judy");
	gpmi_keyword_add_auto(kw, "Bob");

	id = gpmi_keyword_find_int(kw, "John");
	Dprintf("John=%d\n", id);

	str = (char *)gpmi_keyword_find_rev(kw, id);

	if (strcmp(str, "John") != 0)
		fprintf(stderr, "TEST FAILS: gpmi_keyword_find_auto");

	gpmi_keyword_destroy(kw);
}

void test_ptr() {
	gpmi_keyword *kw;
	char *name;

	Dprintf("\n--- Testing pointers and table-list \n");

	/* Check if we can resolve the pointer for table Days  */
	kw = gpmi_keyword_find_ptr(gpmi_shared.gpmi_keywords, "Days");
	Dprintf("kw=%p != NULL\n", kw);

	if (kw==NULL) {
		fprintf(stderr, "TEST FAILS: gpmi_keyword_find_ptr/1");
		return;
	}
	Dprintf("kw->name=%s\n", kw->name);
	if (strcmp(kw->name,"Days")) {
		fprintf(stderr, "TEST FAILS: gpmi_keyword_find_ptr/1");
		return;
	}

	/* Check if table days words */
	name = (char *)gpmi_keyword_find_rev(kw, 2);
	Dprintf("2=%s\n", name);
	if (strcmp(name, "Mon") != 0)
		fprintf(stderr, "TEST FAILS: gpmi_keyword_find_int/2");

}

int main() {
	gpmi_init();
	gpmi_keyword_init();

	test_int();
	test_multi();
	test_str();
	test_auto();
	test_ptr();

	gpmi_keyword_destroy(days);
	gpmi_keyword_destroy(food);
	gpmi_keyword_destroy(acrynoms);

	gpmi_uninit();

	printf("Test finished.\n");
	return 0;
}
